// test.cpp : Defines the entry point for the console application.
//

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <mswsock.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/time.h>
#include <dirent.h>
#include <ctype.h>
#include <malloc.h>

struct addrinfo* aiList;

SOCKET make_server_socket() {
  struct addrinfo aiHints;
  SOCKET s;
  int r;

  aiHints.ai_family=AF_UNSPEC;
  aiHints.ai_family=AF_INET;
  aiHints.ai_socktype=SOCK_STREAM;
  aiHints.ai_protocol=IPPROTO_TCP;
  aiHints.ai_flags=AI_PASSIVE;
  aiHints.ai_addrlen=0;
  aiHints.ai_addr=NULL;
  aiHints.ai_canonname=NULL;
  aiHints.ai_next=NULL;

  if ((r=getaddrinfo(0,"80",&aiHints,&aiList))) {
    fprintf(stderr,"getaddrinfo failed!\n");
    return -1;
  }

  for (; aiList; aiList=aiList->ai_next) {
    s=socket(aiList->ai_family,aiList->ai_socktype,aiList->ai_protocol);
    if (s==-1)
      continue;
    if (bind(s,aiList->ai_addr,aiList->ai_addrlen)) {
      closesocket(s);
      continue;
    }
    if (listen(s,16)) {
      closesocket(s);
      continue;
    }
    break;
  };
  if (!aiList) return -1;
  return s;
}
char* states[]={"WAITFORREAD","WAITFORWRITE","WAITFORTRANSMITFILE"};

enum STATE { WAITFORREAD, WAITFORWRITE, WAITFORTRANSMITFILE };
HANDLE iocp;

struct _state {
  SOCKET sock;
  enum STATE state;
  char inbuf[8192];
  unsigned int ptr;
  unsigned long long tosend;
  char* tofree;
  OVERLAPPED ov;
  HANDLE f;
  int keepalive;
};

static char fromhex(char x) {
  return ((x>='0' && x<='9') ? x-'0' :
	  (x>='a' && x<='f') ? x-'a'+10 :
	  x-'A'+10);
}

static char tohex(char x) {
  return x>9?x+'a':x+'0';
}

static time_t getmtime(const char* dir,const char* file,unsigned long long* filesize,int* isdir) {
  char* x=(char*)alloca(strlen(dir)+strlen(file)+2);
  struct stat ss;
  strcpy(x,dir);
  strcat(x,"/");
  strcat(x,file);
  if (stat(x,&ss)) {
    *filesize=0;
    *isdir=0;
    return 0;
  }
  *filesize=ss.st_size;
  *isdir=S_ISDIR(ss.st_mode);
  return ss.st_mtime;
}

char* dirlisting(const char* dir) {
  char* buf;
  unsigned int size;
  unsigned int off,len;
  char* decdir;
  const char* pdir;
  DIR* d;
  struct dirent* de;
  while (*dir=='/') ++dir;
  off=0;
  decdir=(char*)alloca(strlen(dir)+2);
  {
    int i,j;
    for (i=j=0; dir[i]; ++i,++j) {
      if (dir[i]=='%' && isxdigit(dir[i+1]) && isxdigit(dir[i+2])) {
	if (!(decdir[j]=fromhex(dir[i+1])*16 + fromhex(dir[i+2])))
	  break;
	i+=2;
      } else
	decdir[j]=dir[i];
    }
    decdir[j]=0;
  }
  if (!*dir) {
    pdir=dir;
    strcpy(decdir,".");
  } else
    pdir=decdir;
  if (strstr(decdir,"/.")) return 0;
  buf=(char*)malloc((size=2048+2*(strlen(decdir))));
  if (!buf) return 0;
  if (!(d=opendir(decdir))) { free(buf); return 0; }
  off+=sprintf(buf,"<title>Listing: /%s</title><h1>Listing: /%s</h1><table><tr><th>Name<th>Last Modified<th align=right>Size\n",pdir,pdir);

  while ((de=readdir(d))) {
    if (de->d_name[0]=='.') continue;
    if (size-off<strlen(de->d_name)*3+100) {
      char* x;
      size+=strlen(de->d_name)*3+1024;
      if (!(x=(char*)realloc(buf,size))) {
	free(buf);
	closedir(d);
	return 0;
      }
      buf=x;
    }
    off+=sprintf(buf+off,"<tr><td><a href=\"");
    if (*pdir)
      off+=sprintf(buf+off,"%s/",pdir);
    {
      int i;
      char* x;
      for (x=de->d_name; *x; ++x) {
	if (*x<' ' || *x>='~' || *x=='#' || *x=='"' || *x=='/') {
	  buf[off]='%';
	  buf[off+1]=tohex((*x&0xf0)>>4);
	  buf[off+2]=tohex(*x&0xf);
	  off+=2;
	} else
	  buf[off]=*x;
	++off;
      }
    }
    {
      unsigned long long filesize;
      int isdir;
      time_t tmp=getmtime(decdir,de->d_name,&filesize,&isdir);
      struct tm* t=gmtime(&tmp);
      off+=sprintf(buf+off,"%s\">%s</a><td>",isdir?"/":"",de->d_name);
      if (t)
	off+=strftime(buf+off,size-off,"%Y-%m-%d %H:%M:%S",t);
      else
	off+=sprintf(buf+off,"[invalid]");
      off+=sprintf(buf+off,"<td align=right>%llu\n",filesize);
    }
  }
  off+=sprintf(buf+off,"</table>\n");
  return buf;
}

void handle_http_request(struct _state* st) {
  char* filename,* x;
  int http11=0,ka=0;
  st->keepalive=0;
  st->tofree=0;
  if (st->ptr<sizeof("GET / HTTP/1.0\r\n\r\n")-1) {
    closesocket(st->sock);
    free(st);
    return;
  }
  if (memcmp(st->inbuf,"GET ",4))
err:
  {
    const char* msg="HTTP/1.0 500 unsupported\r\nContent-Type: text/plain\r\n\r\nError parsing the query.\n";
    DWORD written;
    memset(&st->ov,0,sizeof(st->ov));
    st->state=WAITFORWRITE;
    WriteFile((HANDLE)st->sock,msg,strlen(msg),&written,&st->ov);
    return;
  }
  filename=st->inbuf+4;
  for (x=filename; x<st->inbuf+st->ptr; ++x) {
    if (*x=='.' && x[-1]=='/') *x=':';
    if (*x==' ' || *x=='\r') break;
  }
  if (*x!=' ') goto err;
  *x=0; ++x;
  http11=!strncmp(x,"HTTP/1.1",8);
  for (; x<st->inbuf+st->ptr; ++x) {
    for (; x<st->inbuf+st->ptr && *x!='\n'; ++x) ;
    if (!_memicmp(x+1,"connection:",11)) {
      x+=12; while (*x==' ') ++x;
      if (!_memicmp(x,"keep-alive",10))
	ka=1;
      else if (!_memicmp(x,"close",5))
	ka=-1;
      if (ka) break;
    }
    if (ka) break;
  }
  if (!ka)
    ka=http11?1:-1;
  if (ka==-1) ka=0;
  st->keepalive=ka;

  {
    char* x=strrchr(filename,'/');
    if (x && !x[1]) {
      char* dl=dirlisting(filename);
      char* buf;
      unsigned int i;
      if (!dl) goto e404;
      buf=(char*)malloc(strlen(dl)+150);
      if (!buf) {
	free(dl);
	goto e404;
      }
      i=sprintf(buf,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nContent-Length: %u\r\n\r\n%s",strlen(dl),dl);
      free(dl);
      st->tofree=buf;
      DWORD written;
      memset(&st->ov,0,sizeof(st->ov));
      st->state=WAITFORWRITE;
      WriteFile((HANDLE)st->sock,buf,i,&written,&st->ov);	/* if it fails, we can't do anything about it anyway */
      return;
    }
  }
  while (*filename=='/') ++filename;
  if (!*filename) goto err;
  st->f=CreateFileA(filename,FILE_READ_DATA,FILE_SHARE_DELETE|FILE_SHARE_WRITE|FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0);
  if (st->f==INVALID_HANDLE_VALUE)
e404:
  {
    const char* msg="HTTP/1.0 404 not found\r\nContent-Type: text/plain\r\n\r\nFile not found.\n";
    DWORD written;
    memset(&st->ov,0,sizeof(st->ov));
    st->state=WAITFORWRITE;
    WriteFile((HANDLE)st->sock,msg,strlen(msg),&written,&st->ov);	/* if it fails, we can't do anything about it anyway */
    return;
  }
  {
    DWORD a,b;
    unsigned long long c;
    TRANSMIT_FILE_BUFFERS tb;
    struct stat ss;
    char dbuf[50];
    a=GetFileSize(st->f,&b);
    c=(((unsigned long long)b)<<32) | a;
    st->tosend=c;
    if (stat(filename,&ss)==0) {
      struct tm* t;
      t=gmtime(&ss.st_mtime);
      sprintf(dbuf,"time: %d.%d.%d %d:%d:%d\n",t->tm_mday,t->tm_mon+1,t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec);
      strftime(dbuf,sizeof(dbuf),"Last-Modified: %a, %d %b %Y %H:%M:%S GMT\r\n",t);
    } else dbuf[0]=0;
    tb.Head=st->inbuf;
    tb.HeadLength=sprintf(st->inbuf,"HTTP/1.0 200 OK\r\nServer: Kludge/0.1\r\n%sConnection: %s\r\nContent-Type: application/octet-stream\r\nContent-Length: %llu\r\n\r\n",dbuf,ka==1?"keep-alive":"close",c);
    tb.Tail=0;
    tb.TailLength=0;
    memset(&st->ov,0,sizeof(st->ov));
    if (!TransmitFile(st->sock,st->f,st->tosend>0x80000000?0x80000000:(DWORD)st->tosend,0,&st->ov,&tb,TF_USE_KERNEL_APC)) {
      DWORD e=WSAGetLastError();
      if (e != ERROR_IO_PENDING && e!= WSA_IO_PENDING) {
	fprintf(stderr,"TransmitFile failed.\n");
panic:
	closesocket(st->sock);
	CloseHandle(st->f);
	free(st);
	return;
      }
      st->state=WAITFORTRANSMITFILE;
    }
#ifdef CANTHAPPEN
    else {
      fprintf(stderr,"TransmitFile returned early!\n");
      if (st->keepalive) {
        st->state=WAITFORREAD;
	memset(&st->ov,0,sizeof(st->ov));
	if (ReadFile((HANDLE)st->sock,st->inbuf,sizeof(st->inbuf),0,&st->ov)) {
	  fprintf(stderr,"Failed to queue ReadFile!\n");
panic2:
	  closesocket(st->sock);
	  free(st);
	  return;
	}
	if (CreateIoCompletionPort((HANDLE)(st->sock),iocp,(ULONG_PTR)st,0)==0) {
	  fprintf(stderr,"could not queue ReadFile in completion port!\n");
	  goto panic2;
	}
      } else {
	closesocket(st->sock);
	CloseHandle(st->f);
	free(st);
      }
    }
#endif
  }
}

SOCKET s,next;
DWORD nexterror;
OVERLAPPED srv;
struct _state* nextstate;

DWORD WINAPI EventLoopThread(LPVOID ignored) {
  (void)ignored;
  /* now wait for events */
  for (;;) {
    DWORD numberofbytes;
    ULONG_PTR x;
    LPOVERLAPPED o;
    if (GetQueuedCompletionStatus(iocp,&numberofbytes,&x,&o,INFINITE)) {
      struct _state* st;
//      fprintf(stderr,"got event on %d (%d bytes, st=%p)!\n",x==(ULONG_PTR)&s?0:((struct _state*)x)->sock,numberofbytes,x);
      if (x==(ULONG_PTR)&s) {
//	fprintf(stderr,"event on server socket!\n");
	/* incoming connection! */
//	printf("incoming connection with %d bytes to read!\n",numberofbytes);

	st=nextstate;

	nextstate=(struct _state*)malloc(sizeof(*nextstate));
	if (!nextstate) {
	  fprintf(stderr,"malloc() failed!\n");
	  return 1;
	}
	memset(nextstate,0,sizeof(*nextstate));
	nextstate->ptr=0;

	/* queue next overlapping accept */
	next=socket(aiList->ai_family,aiList->ai_socktype,aiList->ai_protocol);
	if (next==-1) {
	  fprintf(stderr,"socket() failed\n");
	  return 1;
	}
	nextstate->sock=next;

	if (!AcceptEx(s,next,nextstate->inbuf,sizeof(nextstate->inbuf)-64,32,32,&nexterror,&srv) && WSAGetLastError()!=ERROR_IO_PENDING) {
	  fprintf(stderr,"AcceptEx() failed, WSAGetLastError()=%d\n",WSAGetLastError());
	  free(st);
	  continue;
	}

	/* queue new socket in completion port */
	if (CreateIoCompletionPort((HANDLE)(st->sock),iocp,(ULONG_PTR)st,0)==0) {
	  fprintf(stderr,"Failed to queue AcceptEx in completion port (%d)!\n",GetLastError());
	  closesocket(st->sock);
	  free(st);
	  continue;
	}
	st->state=WAITFORREAD;
	goto readevent;
      } else {
	/* event on some socket */
	st=(struct _state*)x;
//	fprintf(stderr,"event on socket %d (state %s): %lu bytes\n",st->sock,states[st->state],numberofbytes);
	if (st->state==WAITFORREAD)
readevent:
	{
	  DWORD i,x;
	  if (numberofbytes+st->ptr>=sizeof(st->inbuf)) {
	    fprintf(stderr,"buffer overflow!\n");
	    return 1;
	  }
	  x=0;
	  st->ptr+=numberofbytes;
	  for (i=0; i+3<=st->ptr; ++i)
	    if (st->inbuf[i]=='\r' && st->inbuf[i+1]=='\n' && st->inbuf[i+2]=='\r' && st->inbuf[i+3]=='\n') {
	      x=1;
	      break;
	    }
	  if (x) {
//	    fprintf(stderr,"got full HTTP header!\n");
	    handle_http_request(st);
	  } else {
	    if (st->ptr>=sizeof(st->inbuf)) {
	      fprintf(stderr,"Read 8192 bytes but no end of the header is in sight...\n");
	      closesocket(st->sock);
	      free(st);
	    } else if (ReadFile((HANDLE)st->sock,st->inbuf+st->ptr,sizeof(st->inbuf)-st->ptr,0,&st->ov) && GetLastError() != ERROR_IO_PENDING) {
	      fprintf(stderr,"Failed to queue ReadFile (error %d)!\n",GetLastError());
	      closesocket(st->sock);
	      free(st);
	    }
	    continue;
	  }
	} else if (st->state==WAITFORTRANSMITFILE) {
	  /* we might have to send more */
	  if (st->tosend>numberofbytes) {
	    /* have to send more */
	    st->tosend-=numberofbytes;
	    TRANSMIT_FILE_BUFFERS tb;
	    memset(&tb,0,sizeof(tb));
	    if (!TransmitFile(st->sock,st->f,st->tosend>0x80000000?0x80000000:(DWORD)st->tosend,0,&st->ov,&tb,TF_USE_KERNEL_APC)) {
	      DWORD e=WSAGetLastError();
	      if (e != ERROR_IO_PENDING && e!= WSA_IO_PENDING) {
		fprintf(stderr,"TransmitFile failed.\n");
panic:
		CloseHandle(st->f);
		goto panic2;
	      }
	      if (CreateIoCompletionPort((HANDLE)(st->sock),iocp,(ULONG_PTR)st,0)==0) {
		fprintf(stderr,"Failed to queue TransmitFile in completion port!\n");
		goto panic;
	      }
	    }
	  } else {
	    CloseHandle(st->f);
	    if (st->keepalive) {
	      memset(&st->ov,0,sizeof(st->ov));
//	      fprintf(stderr,"Keepalive!  Queueing ReadFile...\n");
	      st->state=WAITFORREAD;
	      if (ReadFile((HANDLE)st->sock,st->inbuf,sizeof(st->inbuf),0,&st->ov)) {
		fprintf(stderr,"Overlapped ReadFile succeeded unexpectedly!\n");
		goto readevent;
	      }
	      if (GetLastError() != ERROR_IO_PENDING) {
		fprintf(stderr,"Read error!\n");
panic2:
		closesocket(st->sock);
		free(st);
		return 1;
	      }
	    } else {
	      closesocket(st->sock);
	      free(st);
	    }
	  }
	} else {
	  if (st->tofree) free(st->tofree);
	  closesocket(st->sock);
	  free(st);
	}
      }
    }
  }
}

int main(int argc,char* argv[]) {
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2,0),&wsaData)) {
    fprintf(stderr,"no suitable winsock version found!\n");
    return 2;
  }
  s=make_server_socket();
  if (s==-1) {
    fprintf(stderr,"BIND failed!\n");
    return 2;
  }
  fprintf(stderr,"Bound to port 80.\n");
  next=socket(aiList->ai_family,aiList->ai_socktype,aiList->ai_protocol);
  if (next==-1) {
    fprintf(stderr,"socket() failed\n");
    return 1;
  }
  if (!(nextstate=(struct _state*)malloc(sizeof(*nextstate)))) {
    fprintf(stderr,"malloc failed!\n");
    return 1;
  }
  memset(nextstate,0,sizeof(*nextstate));
  nextstate->state=WAITFORREAD;
  nextstate->sock=next;
  if (!AcceptEx(s,next,nextstate->inbuf,sizeof(nextstate->inbuf)-64,32,32,&nexterror,&srv) && WSAGetLastError()!=ERROR_IO_PENDING) {
    fprintf(stderr,"accept failed, nexterror=%d (WSAGetLastError=%d)\n",nexterror,WSAGetLastError());
    return 1;
  }
  iocp=CreateIoCompletionPort(INVALID_HANDLE_VALUE,NULL,0,0);
  if (!iocp) {
    fprintf(stderr,"completion ports not supported\n");
    return 1;
  }
  if (CreateIoCompletionPort((HANDLE)s,iocp,(ULONG_PTR)&s,0)==0) {
    fprintf(stderr,"could not queue s in completion port!\n");
    return 1;
  }

  if (argc>1) {
    unsigned int threads;
    if (sscanf(argv[1],"%u",&threads)==1) {
      if (threads>100) threads=100;
      if (threads>1) {
	HANDLE t[100];
	unsigned int i;
	for (i=0; i<threads; ++i) {
	  if (!(t[i]=CreateThread(NULL,8192,EventLoopThread,NULL,0,NULL))) {
	    fprintf(stderr,"Thread creation failed, error code %d\n",GetLastError());
	    return 1;
	  }
	  fprintf(stderr,"created thread %d\n",i);
	}
	WaitForMultipleObjects(threads,t,TRUE,INFINITE);
      }
    }
  }
  EventLoopThread(NULL);
}
